import json
import os
import sys
import time

from subprocess import run

from .compat import Compat, Version
from .state import STATE
from .find_julia import julia_version, find_julia
from .install_julia import log

### META

META_VERSION = 1 # increment whenever the format changes

def load_meta():
    fn = STATE['meta']
    if os.path.exists(fn):
        with open(fn) as fp:
            meta = json.load(fp)
            if meta.get('meta_version') == META_VERSION:
                return meta

def save_meta(meta):
    assert isinstance(meta, dict)
    assert meta.get('meta_version') == META_VERSION
    fn = STATE['meta']
    with open(fn, 'w') as fp:
        json.dump(meta, fp)

### RESOLVE

class PkgSpec:
    def __init__(self, name, uuid, dev=False, version=None, path=None, url=None, rev=None):
        self.name = name
        self.uuid = uuid
        self.dev = dev
        self.version = version
        self.path = path
        self.url = url
        self.rev = rev

    def jlstr(self):
        args = ['name="{}"'.format(self.name), 'uuid="{}"'.format(self.uuid)]
        if self.path is not None:
            args.append('path=raw"{}"'.format(self.path))
        if self.url is not None:
            args.append('url=raw"{}"'.format(self.url))
        if self.rev is not None:
            args.append('rev=raw"{}"'.format(self.rev))
        return "Pkg.PackageSpec({})".format(', '.join(args))

    def dict(self):
        ans = {
            "name": self.name,
            "uuid": self.uuid,
            "dev": self.dev,
            "version": str(self.version),
            "path": self.path,
            "url": self.url,
            "rev": self.rev,
        }
        return {k:v for (k,v) in ans.items() if v is not None}

    def depsdict(self):
        ans = {}
        ans['uuid'] = self.uuid
        if self.dev:
            ans['dev'] = self.dev
        if self.version is not None:
            ans['version'] = str(self.version)
        if self.path is not None:
            ans['path'] = self.path
        if self.url is not None:
            ans['url'] = self.url
        if self.rev is not None:
            ans['rev'] = self.rev
        return ans

def can_skip_resolve():
    # resolve if we haven't resolved before
    deps = load_meta()
    if deps is None:
        return False
    # resolve whenever Julia changes
    exe = deps.get("executable")
    if exe is None:
        return False
    ver = deps.get("version")
    exever = julia_version(exe)
    if ver is None or exever is None or ver != str(exever):
        return False
    # resolve whenever swapping between dev/not dev
    isdev = deps.get("dev")
    if isdev is None or isdev != STATE["dev"]:
        return False
    # resolve whenever anything in sys.path changes
    timestamp = deps.get("timestamp")
    if timestamp is None:
        return False
    timestamp = max(os.path.getmtime(STATE["meta"]), timestamp)
    sys_path = deps.get("sys_path")
    if sys_path is None or sys_path != sys.path:
        return False
    for path in sys.path:
        if not path:
            path = os.getcwd()
        if not os.path.exists(path):
            continue
        if os.path.getmtime(path) > timestamp:
            return False
        if os.path.isdir(path):
            fn = os.path.join(path, "juliapkg.json")
            if os.path.exists(fn) and os.path.getmtime(fn) > timestamp:
                return False
    return deps

def deps_files():
    ans = []
    # look in sys.path
    for path in sys.path:
        if not path:
            path = os.getcwd()
        if not os.path.isdir(path):
            continue
        fn = os.path.join(path, "juliapkg.json")
        ans.append(fn)
        for subdir in os.listdir(path):
            fn = os.path.join(path, subdir, "juliapkg.json")
            ans.append(fn)
    # {prefix}/juliapkg.json
    if STATE['prefix']:
        fn = os.path.join(STATE['prefix'], 'juliapkg.json')
        ans.append(fn)
    # ~/.pyjuliapkg.json
    fn = os.path.join(os.path.expanduser('~'), '.pyjuliapkg.json')
    ans.append(fn)
    return list(set(fn for fn in ans if os.path.isfile(fn)))

def required_packages():
    # read all dependencies into a dict: name -> key -> file -> value
    import json
    all_deps = {}
    for fn in deps_files():
        with open(fn) as fp:
            deps = json.load(fp)
        for (name, kvs) in deps.get("packages", {}).items():
            dep = all_deps.setdefault(name, {})
            for (k, v) in kvs.items():
                if k == 'path':
                    # resolve paths relative to the directory containing the file
                    v = os.path.join(os.path.dirname(fn), v)
                dep.setdefault(k, {})[fn] = v
    # merges non-unique values
    def merge_unique(dep, kfvs, k):
        fvs = kfvs.pop(k, None)
        if fvs is not None:
            vs = set(fvs.values())
            if len(vs) == 1:
                dep[k], = vs
            elif vs:
                raise Exception("'{}' entries are not unique:\n{}".format(k, '\n'.join(['- {!r} at {}'.format(v,f) for (f,v) in fvs.items()])))
    # merges compat entries
    def merge_compat(dep, kfvs, k):
        fvs = kfvs.pop(k, None)
        if fvs is not None:
            compats = list(map(Compat.parse, fvs.values()))
            compat = compats[0]
            for c in compats[1:]:
                compat &= c
            if not compat:
                raise Exception("'{}' entries have empty intersection:\n{}".format(k, '\n'.join(['- {!r} at {}'.format(v,f) for (f,v) in fvs.items()])))
            else:
                dep[k] = str(compat)
    # merges booleans with any
    def merge_any(dep, kfvs, k):
        fvs = kfvs.pop(k, None)
        if fvs is not None:
            dep[k] = any(fvs.values())
    # merge dependencies: name -> key -> value
    deps = []
    for (name, kfvs) in all_deps.items():
        kw = {'name': name}
        merge_unique(kw, kfvs, 'uuid')
        merge_unique(kw, kfvs, 'path')
        merge_unique(kw, kfvs, 'url')
        merge_unique(kw, kfvs, 'rev')
        merge_compat(kw, kfvs, 'version')
        merge_any(kw, kfvs, 'dev')
        deps.append(PkgSpec(**kw))
    return deps

def required_julia():
    import json
    compats = {}
    for fn in deps_files():
        with open(fn) as fp:
            deps = json.load(fp)
            c = deps.get("julia")
            if c is not None:
                compats[fn] = Compat.parse(c)
    compat = None
    for c in compats.values():
        if compat is None:
            compat = c
        else:
            compat &= c
    if compat is not None and not compat:
        raise Exception("'julia' compat entries have empty intersection:\n{}".format('\n'.join(['- {!r} at {}'.format(v,f) for (f,v) in compats.items()])))
    return compat

def resolve(force=False, dry_run=False):
    # see if we can skip resolving
    if not force:
        if STATE['resolved']:
            return True
        deps = can_skip_resolve()
        if deps:
            STATE['resolved'] = True
            STATE['executable'] = deps['executable']
            STATE['version'] = Version(deps['version'])
            return True
    if dry_run:
        return False
    STATE['resolved'] = False
    # get julia compat
    compat = required_julia()
    # get required packages
    pkgs = required_packages()
    # find a compatible julia executable
    log(f'Locating Julia{"" if compat is None else " "+str(compat)}')
    exe, ver = find_julia(compat=compat, prefix=STATE['install'], install=True, upgrade=True)
    log(f'Using Julia {ver} at {exe}')
    # set up the project
    project = STATE['project']
    log(f'Using Julia project at {project}')
    os.makedirs(project, exist_ok=True)
    # write a Project.toml specifying UUIDs and compatibility of required packages
    with open(os.path.join(project, "Project.toml"), "wt") as fp:
        print('[deps]', file=fp)
        for pkg in pkgs:
            print(f'{pkg.name} = "{pkg.uuid}"', file=fp)
        print(file=fp)
        print('[compat]', file=fp)
        for pkg in pkgs:
            if pkg.version:
                print(f'{pkg.name} = "{pkg.version}"', file=fp)
        print(file=fp)
    # install the packages
    dev_pkgs = ', '.join([pkg.jlstr() for pkg in pkgs if pkg.dev])
    add_pkgs = ', '.join([pkg.jlstr() for pkg in pkgs if not pkg.dev])
    script = ['import Pkg']
    if dev_pkgs:
        script.append(f'Pkg.develop([{dev_pkgs}])')
    if add_pkgs:
        script.append(f'Pkg.add([{add_pkgs}])')
    script.append('Pkg.resolve()')
    log(f'Installing packages:')
    for line in script:
        log('julia>', line, cont=True)
    run([exe, '--project='+project, '-e', '; '.join(script)], check=True)
    # record that we resolved
    save_meta({
        "meta_version": META_VERSION,
        "dev": STATE["dev"],
        "version": str(ver),
        "executable": exe,
        "timestamp": time.time(),
        "sys_path": sys.path,
        "pkgs": [pkg.dict() for pkg in pkgs],
    })
    STATE['resolved'] = True
    STATE['executable'] = exe
    STATE['version'] = ver
    return True

def executable():
    resolve()
    return STATE['executable']

def project():
    resolve()
    return STATE['project']

def cur_deps_file():
    if STATE['prefix']:
        return os.path.join(STATE['prefix'], 'juliapkg.json')
    else:
        return os.path.join(os.path.expanduser('~'), '.pyjuliapkg.json')

def load_cur_deps():
    fn = cur_deps_file()
    if os.path.exists(fn):
        with open(fn) as fp:
            deps = json.load(fp)
    else:
        deps = {}
    return deps

def write_cur_deps(deps):
    fn = cur_deps_file()
    if deps:
        with open(fn, 'w') as fp:
            json.dump(deps, fp)
    else:
        if os.path.exists(fn):
            os.remove(fn)

def status():
    res = resolve(dry_run=True)
    print('JuliaPkg Status')
    fn = cur_deps_file()
    if os.path.exists(fn):
        with open(fn) as fp:
            deps = json.load(fp)
    else:
        deps = {}
    st = '' if deps else ' (empty project)'
    print(f'{fn}{st}')
    if res:
        exe = STATE['executable']
        ver = STATE['version']
    else:
        print('Not resolved (resolve for more information)')
    jl = deps.get('julia')
    if res or jl:
        print('Julia', end='')
        if res:
            print(f' {ver}', end='')
        if jl:
            print(f' ({jl})', end='')
        if res:
            print(f' @ {exe}', end='')
        print()
    pkgs = deps.get('packages')
    if pkgs:
        print(f'Packages:')
        for (name, info) in pkgs.items():
            print(f'  {name}: {info}')

def set_julia_compat(compat):
    deps = load_cur_deps()
    if compat is None:
        if 'julia' in deps:
            del deps['julia']
    else:
        if isinstance(compat, str):
            compat = Compat.parse(compat)
        elif not isinstance(compat, Compat):
            raise TypeError
        deps['julia'] = str(compat)
    write_cur_deps(deps)
    STATE['resolved'] = False

def add(pkg, *args, **kwargs):
    deps = load_cur_deps()
    _add(deps, pkg, *args, **kwargs)
    write_cur_deps(deps)
    STATE['resolved'] = False

def _add(deps, pkg, uuid=None, **kwargs):
    if isinstance(pkg, PkgSpec):
        pkgs = deps.setdefault('packages', {})
        pkgs[pkg.name] = pkg.depsdict()
    elif isinstance(pkg, str):
        if uuid is None:
            raise TypeError('uuid is required')
        pkg = PkgSpec(pkg, uuid, **kwargs)
        _add(deps, pkg)
    else:
        for p in pkg:
            _add(deps, p)

def rm(pkg):
    deps = load_cur_deps()
    _rm(deps, pkg)
    write_cur_deps(deps)
    STATE['resolved'] = False

def _rm(deps, pkg):
    if isinstance(pkg, PkgSpec):
        _rm(deps, pkg.name)
    elif isinstance(pkg, str):
        pkgs = deps.setdefault('packages', {})
        if pkg in pkgs:
            del pkgs[pkg]
        if not pkgs:
            del deps['packages']
    else:
        for p in pkg:
            _rm(deps, p)



import re
import semantic_version as sv


Version = sv.Version


class Compat:
    """A Julia compat specifier."""

    re_range = re.compile(r'^([~^])?([0-9]+)(?:\.([0-9]+)(?:\.([0-9]+))?)?$')

    def __init__(self, clauses=[]):
        self.clauses = list(clauses)

    def __str__(self):
        return ', '.join(str(clause) for clause in self.clauses)

    def __repr__(self):
        return f'{type(self).__name__}({self.clauses!r})'

    def __contains__(self, v):
        return any(v in clause for clause in self.clauses)

    def __and__(self, other):
        clauses = []
        for clause1 in self.clauses:
            for clause2 in other.clauses:
                clause = clause1 & clause2
                if clause is not None:
                    clauses.append(clause)
        return Compat(clauses)

    def __bool__(self):
        return bool(self.clauses)

    @classmethod
    def parse(cls, verstr):
        """Parse a Julia compat specifier from a string.

        A specifier is a comma-separated list of clauses. The prefixes '^', '~' and '=='
        are supported. No prefix is equivalent to '^'.
        """
        clauses = []
        if verstr.strip():
            for part in verstr.split(','):
                part = part.strip()
                m = cls.re_range.match(part)
                if m is not None:
                    prefix, major, minor, patch = m.groups()
                    prefix = prefix or '^'
                    assert prefix in ('^', '~')
                    major = int(major)
                    minor = None if minor is None else int(minor)
                    patch = None if patch is None else int(patch)
                    version = Version(major=major, minor=minor or 0, patch=patch or 0)
                    if prefix == '^':
                        if major != 0 or minor is None:
                            nfixed = 1
                        elif minor != 0 or patch is None:
                            nfixed = 2
                        else:
                            nfixed = 3
                    elif prefix == '~':
                        if minor is None:
                            nfixed = 1
                        else:
                            nfixed = 2
                    clause = Range(version, nfixed)
                elif part.startswith('=='):
                    version = Version(part[2:])
                    clause = Eq(version)
                else:
                    raise ValueError(f'invalid version: {part!r}')
                clauses.append(clause)
        return cls(clauses)


class Eq:

    def __init__(self, version):
        self.version = version

    def __str__(self):
        return f'=={self.version}'

    def __repr__(self):
        return f'{type(self).__name__}({self.version!r})'

    def __contains__(self, v):
        if isinstance(v, Version):
            return v == self.version
        return False

    def __and__(self, other):
        if self.version in other:
            return self


class Range:

    def __init__(self, version, nfixed):
        self.version = version
        self.nfixed = nfixed

    def __str__(self):
        v = self.version
        n = self.nfixed
        if n == 1:
            if v.major != 0:
                if v.patch != 0:
                    return f'^{v.major}.{v.minor}.{v.patch}'
                elif v.minor != 0:
                    return f'^{v.major}.{v.minor}'
                else:
                    return f'^{v.major}'
            elif v.minor == 0 and v.patch == 0:
                return f'^0'
            else:
                assert False
        elif n == 2:
            if v.major == 0:
                if v.minor == 0:
                    if v.patch == 0:
                        return f'~0.0'
                    else:
                        return f'~0.0.{v.patch}'
                else:
                    if v.patch == 0:
                        return f'^0.{v.minor}'
                    else:
                        return f'^0.{v.minor}.{v.patch}'
            else:
                if v.patch == 0:
                    return f'~{v.major}.{v.minor}'
                else:
                    return f'~{v.major}.{v.minor}.{v.patch}'
        elif n == 3:
            if v.major == 0 and v.minor == 0:
                return f'^0.0.{v.patch}'
            else:
                assert False
        else:
            assert False

    def __repr__(self):
        return f'{type(self).__name__}({self.version!r}, {self.nfixed!r})'

    def __contains__(self, v):
        if isinstance(v, Version):
            n = self.nfixed
            v0 = self.version
            return (n < 1 or v.major == v0.major) and (n < 2 or v.minor == v0.minor) and (n < 3 or v.patch == v0.patch) and v >= v0

    def __and__(self, other):
        if isinstance(other, Range):
            n0 = self.nfixed
            v0 = self.version
            n1 = other.nfixed
            v1 = other.version
            nmin = min(n0, n1)
            nmax = max(n0, n1)
            vmax = max(v0, v1)
            if (nmin < 1 or v0.major == v1.major) and (nmin < 2 or v0.minor == v1.minor) and (nmin < 3 or v0.patch == v1.patch):
                if vmax in self and vmax in other:
                    return Range(vmax, nmax)
        elif isinstance(other, Eq):
            if other.version in self:
                return other
        else:
            return NotImplemented



import os

STATE = {}

def reset_state():
    global STATE
    STATE = {}

    # Are we running a dev version?
    STATE['dev'] = os.path.exists(os.path.join(os.path.dirname(__file__), '..', '..', 'pyproject.toml'))

    # Find the Julia depot
    depot_path = os.getenv('JULIA_DEPOT_PATH')
    if depot_path:
        sep = ';' if os.name == 'nt' else ':'
        STATE['depot'] = os.path.abspath(depot_path.split(sep)[0])
    else:
        STATE['depot'] = os.path.abspath(os.path.join(os.path.expanduser('~'), '.julia'))

    # Determine where to put the julia environment
    # TODO: Can we more direcly figure out the environment from which python was called? Maybe find the first PATH entry containing python?
    vprefix = os.getenv('VIRTUAL_ENV')
    cprefix = os.getenv('CONDA_PREFIX')
    if cprefix and vprefix:
        raise Exception('You are using both a virtual and conda environment, cannot figure out which to use!')
    elif cprefix:
        prefix = cprefix
    elif vprefix:
        prefix = vprefix
    else:
        prefix = None
    STATE['prefix'] = prefix
    if prefix is None:
        STATE['project'] = os.path.join(STATE['depot'], 'environments', 'pyjuliapkg')
    else:
        STATE['project'] = os.path.abspath(os.path.join(prefix, 'julia_env'))

    # meta file
    STATE['meta'] = os.path.join(STATE['project'], 'pyjuliapkgmeta.json')
    STATE['install'] = os.path.join(STATE['project'], 'install')

    # resolution
    STATE['resolved'] = False

reset_state()



import hashlib
import io
import gzip
import json
import os
import platform
import shutil
import subprocess
import tarfile
import tempfile
import time
import urllib.request
import warnings
import zipfile

from .compat import Version

_all_julia_versions = None
_julia_versions_url = 'https://julialang-s3.julialang.org/bin/versions.json'

def log(*args, cont=False):
    prefix = '          ' if cont else '[juliapkg]'
    print(prefix, *args)

def all_julia_versions():
    global _all_julia_versions
    if _all_julia_versions is None:
        url = _julia_versions_url
        log(f'Querying Julia versions from {url}')
        with urllib.request.urlopen(url) as fp:
            _all_julia_versions = json.load(fp)
    return _all_julia_versions

os_aliases = {
    'darwin': 'mac',
    'windows': 'winnt',
}

def get_os():
    os = platform.system().lower()
    return os_aliases.get(os.lower(), os)

arch_aliases = {
    'arm64': 'aarch64',
    'i386': 'i686',
    'amd64': 'x86_64',
}

def get_arch():
    arch = platform.machine().lower()
    return arch_aliases.get(arch.lower(), arch)

libc_aliases = {
    'glibc': 'gnu',
}

def get_libc():
    libc = platform.libc_ver()[0].lower()
    return libc_aliases.get(libc, libc)

def compatible_julia_versions(compat=None):
    os = get_os()
    arch = get_arch()
    libc = get_libc()
    if libc == '' and os == 'linux':
        warnings.warn('could not determine libc version - assuming glibc')
        libc = 'gnu'
    ans = {}
    for (k, v) in all_julia_versions().items():
        v = v.copy()
        if not v['stable']:
            continue
        files = []
        for f in v['files']:
            assert f['version'] == k
            if not any(f['url'].endswith(ext) for ext in julia_installers):
                continue
            if f['os'] != os:
                continue
            if f['arch'] != arch:
                continue
            if os == 'linux' and f['triplet'].split('-')[2] != libc:
                continue
            if compat is not None:
                try:
                    ver = Version(f['version'])
                except Exception:
                    continue
                if ver not in compat:
                    continue
            files.append(f)
        if not files:
            continue
        v['files'] = files
        ans[k] = v
    triplets = {f['triplet'] for (k, v) in ans.items() for f in v['files']}
    if len(triplets) > 1:
        raise Exception(f'multiple matching triplets {sorted(triplets)} - this is probably a bug, please report')
    return ans

def best_julia_version(compat=None):
    vers = compatible_julia_versions(compat)
    if not vers:
        raise Exception(f'no version of Julia is compatible with {compat}')
    v = sorted(vers.keys(), key=Version, reverse=True)[0]
    return v, vers[v]

def install_julia(ver, prefix):
    for f in ver['files']:
        url = f['url']
        # find a suitable installer
        installer = None
        for ext in julia_installers:
            if url.endswith(ext):
                installer = julia_installers[ext]
                break
        if installer is None:
            continue
        # download julia
        buf = download_julia(f)
        # include the version in the prefix
        v = f['version']
        log(f'Installing Julia {v} to {prefix}')
        if os.path.exists(prefix):
            shutil.rmtree(prefix)
        if os.path.dirname(prefix):
            os.makedirs(os.path.dirname(prefix), exist_ok=True)
        installer(f, buf, prefix)
        return
    raise Exception('no installable Julia version found')

def download_julia(f):
    url = f['url']
    sha256 = f['sha256']
    size = f['size']
    log(f'Downloading Julia from {url}')
    buf = io.BytesIO()
    freq = 5
    t = time.time() + freq
    with urllib.request.urlopen(url) as f:
        while True:
            data = f.read(1<<16)
            if not data:
                break
            buf.write(data)
            if time.time() > t:
                log(f'  downloaded {buf.tell()/(1<<20):.1f} MB of {size/(1<<20):.1f} MB', cont=True)
                t = time.time() + freq
    log('  download complete', cont=True)
    log(f'Verifying download')
    buf.seek(0)
    m = hashlib.sha256()
    m.update(buf.read())
    sha256actual = m.hexdigest()
    if sha256actual != sha256:
        raise Exception(f'SHA-256 hash does not match, got {sha256actual}, expecting {sha256}')
    buf.seek(0)
    return buf

def install_julia_zip(f, buf, prefix):
    with tempfile.TemporaryDirectory() as tmpdir:
        # extract all files
        with zipfile.ZipFile(buf) as zf:
            zf.extractall(tmpdir)
        # copy stuff out
        srcdirs = [d for d in os.listdir(tmpdir) if d.startswith('julia')]
        if len(srcdirs) != 1:
            raise Exception('expecting one julia* directory')
        shutil.copytree(os.path.join(tmpdir, srcdirs[0]), prefix, symlinks=True)

def install_julia_tar_gz(f, buf, prefix):
    with tempfile.TemporaryDirectory() as tmpdir:
        # extract all files
        with gzip.GzipFile(fileobj=buf) as gf:
            with tarfile.TarFile(fileobj=gf) as tf:
                tf.extractall(tmpdir)
        # copy stuff out
        srcdirs = [d for d in os.listdir(tmpdir) if d.startswith('julia')]
        if len(srcdirs) != 1:
            raise Exception('expecting one julia* directory')
        shutil.copytree(os.path.join(tmpdir, srcdirs[0]), prefix, symlinks=True)

def install_julia_dmg(f, buf, prefix):
    with tempfile.TemporaryDirectory() as tmpdir:
        # write the dmg file out
        dmg = os.path.join(tmpdir, 'dmg')
        with open(dmg, 'wb') as f:
            f.write(buf.read())
        # mount it
        mount = os.path.join(tmpdir, 'mount')
        subprocess.run(['hdiutil', 'mount', '-mount', 'required', '-mountpoint', mount, dmg], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        try:
            # copy stuff out
            appdirs = [d for d in os.listdir(mount) if d.startswith('Julia') and d.endswith('.app')]
            if len(appdirs) != 1:
                raise Exception('expecting one Julia*.app directory')
            srcdir = os.path.join(mount, appdirs[0], 'Contents', 'Resources', 'julia')
            shutil.copytree(srcdir, prefix, symlinks=True)
        finally:
            # unmount
            subprocess.run(['umount', mount], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

julia_installers = {
    '.tar.gz': install_julia_tar_gz,
    '.zip': install_julia_zip,
    '.dmg': install_julia_dmg,
}



import json
import os
import shutil

from subprocess import run
from .install_julia import best_julia_version, install_julia, log
from .compat import Version, Compat
from .state import STATE

def julia_version(exe):
    try:
        words = run([exe, '--version'], check=True, capture_output=True, encoding='utf8').stdout.strip().split()
        if words[0].lower() == 'julia' and words[1].lower() == 'version':
            return Version(words[2])
    except:
        pass

def find_julia(compat=None, prefix=None, install=False, upgrade=False):
    """Find a Julia executable compatible with compat.

    Args:
        compat: A juliapkg.compat.Compat giving bounds on the allowed version of Julia.
        prefix: An optional prefix in which to look for or install Julia.
        install: If True, install Julia if it is not found. This will use JuliaUp if
            available, otherwise will install into the given prefix.
        upgrade: If True, find the latest compatible release. Implies install=True.

    As a special case, upgrade=True does not apply when Julia is found in the PATH, because
    if it is already installed then the user is already managing their own Julia versions.
    """
    # if upgrading, fix compat to the best version
    orig_compat = compat
    if upgrade:
        verstr, _ = best_julia_version(compat)
        compat = Compat.parse('=='+verstr)
        install = True
    # first look in the prefix
    if prefix is not None:
        ext = '.exe' if os.name == 'nt' else ''
        pr_exe = shutil.which(os.path.join(prefix, 'bin', 'julia' + ext))
        pr_ver = julia_version(pr_exe)
        if pr_ver is not None:
            if compat is None or pr_ver in compat:
                return (pr_exe, pr_ver)
    # see if juliaup is installed
    ju_exe = shutil.which('juliaup')
    if ju_exe:
        ans = ju_find_julia(compat, install=install)
        if ans:
            return ans
    else:
        # see if julia is installed
        jl_exe = shutil.which('julia')
        jl_ver = julia_version(jl_exe)
        if jl_ver is not None:
            if orig_compat is None or jl_ver in orig_compat:
                return (jl_exe, jl_ver)
            else:
                log('WARNING: You have Julia installed but it is not compatible with your juliapkg dependencies.')
                log('  It is recommended that you upgrade Julia or install JuliaUp.')
    # install into the prefix
    if install and prefix is not None:
        ver, info = best_julia_version(compat)
        log(f'WARNING: About to install Julia to {prefix}.')
        log(f'  If you use juliapkg in more than one environment, you are likely to have Julia')
        log(f'  installed in multiple locations. It is recommended to install JuliaUp')
        log(f'  (https://github.com/JuliaLang/juliaup) or Julia (https://julialang.org/downloads)')
        log(f'  yourself.')
        install_julia(info, prefix)
        pr_exe = shutil.which(os.path.join(prefix, 'bin', 'julia' + ext))
        pr_ver = julia_version(pr_exe)
        if pr_ver is not None:
            if compat is None or pr_ver in compat:
                return (pr_exe, pr_ver)
        assert False
    # failed
    compatstr = '' if orig_compat is None else f' {orig_compat}'
    raise Exception(f'could not find Julia{compatstr}')

def ju_find_julia(compat=None, install=False):
    # see if it is already installed
    ans = ju_find_julia_noinstall(compat)
    if ans:
        return ans
    # install it
    if install:
        ver, _ = best_julia_version(compat)
        run(['juliaup', 'add', ver], check=True)
        ans = ju_find_julia_noinstall(compat)
        if ans:
            return ans
        Exception(f'JuliaUp just installed Julia {ver} but cannot find it')

def ju_find_julia_noinstall(compat=None):
    judir = os.path.join(STATE['depot'], 'juliaup')
    metaname = os.path.join(judir, 'juliaup.json')
    if os.path.exists(metaname):
        with open(metaname) as fp:
            meta = json.load(fp)
        versions = []
        for (verstr, info) in meta.get('InstalledVersions', {}).items():
            ver = Version(verstr.split('~')[0])
            ver = Version(major=ver.major, minor=ver.minor, patch=ver.patch, prerelease=ver.prerelease, build=tuple(x for x in ver.build if x != '0'))
            if compat is None or ver in compat:
                if 'Path' in info:
                    ext = '.exe' if os.name == 'nt' else ''
                    exe = os.path.abspath(os.path.join(judir, info['Path'], 'bin', 'julia'+ext))
                    versions.append((exe, ver))
        versions.sort(key=lambda x: x[1], reverse=True)
        for (exe, _) in versions:
            ver = julia_version(exe)
            if compat is None or ver in compat:
                return (exe, ver)
